from datetime import datetime as dt
from functools import wraps

from pydantic import StrictFloat, StrictInt, StrictStr, conlist, create_model

DATATYPE_MAP = {'string': StrictStr,
                'bigint': StrictInt,
                'integer': StrictInt,
                'double': StrictFloat,
                'array<string>': conlist(StrictStr, min_items=1)}


def timeit(func):
    """
    Timing function to time error collection.
    Args:
        func:

    Returns:
    """

    @wraps(func)
    def timed(*args, **kw):
        tstart = dt.now()
        output = func(*args, **kw)
        tend = dt.now()
        print('"{}" took {} to execute.\n'.format(func.__name__, (tend - tstart)))
        return output

    return timed


def list_to_set_in_dict(obj):
    """
    Recursively convert list values inside a dic to sets
    Args:
        obj: Current value being checked

    Returns:
        obj: Dict with lists converted to sets.
    """
    if isinstance(obj, list):
        return {list_to_set_in_dict(v) for v in obj}
    elif isinstance(obj, dict):
        return {k: list_to_set_in_dict(v) for k, v in obj.items()}
    return obj


def lower_and_strip(obj):
    """
    Lower and strip all values in a list
    Args:
        obj:

    Returns:

    """
    for key, val in obj.items():
        if isinstance(val, list):
            obj[key] = [v.lower().strip() for v in val]
        elif isinstance(val, dict):
            obj[key] = lower_and_strip(obj[key])
    return obj


def chunks(lst, n):
    """Yield successive n-sized chunks from lst."""
    for i in range(0, len(lst), n):
        yield lst[i:i + n]


def build_validation_schema_sample_level(desc_df, validators=None):
    """
    Takes a description DF of a sample and returns a pydantic class with the values
    Args:
        validators:
        desc_df: Result of "DESCRIBE repo.samples"

    Returns:

    """
    curated_fields = desc_df[desc_df['column_name'].str.startswith('curated')]
    schema_dict = dict(zip(curated_fields['column_name'], curated_fields['column_type']))
    for k, v in schema_dict.items():
        schema_dict[k] = (DATATYPE_MAP[v], ...)
    if validators is not None:
        schema = create_model('Schema', **schema_dict, __validators__=validators)
    else:
        schema = create_model('Schema', **schema_dict)
    return schema


def build_validation_schema_dataset_level(desc_df, validators=None):
    """
    Takes a description DF of a sample and returns a pydantic class with the values
    Args:
        validators:
        desc_df: Result of "DESCRIBE repo.samples"

    Returns:

    """
    # curated_fields = desc_df[desc_df['column_name'].str.startswith('curated')]
    schema_dict = dict(zip(desc_df['column_name'], desc_df['column_type']))
    for k, v in schema_dict.items():
        schema_dict[k] = (DATATYPE_MAP[v], ...)
    if validators is not None:
        schema = create_model('Schema', **schema_dict, __validators__=validators)
    else:
        schema = create_model('Schema', **schema_dict)
    return schema


def split_ontological_fields_by_type(ontological_fields, desc_df):
    """
    Given a list of ontological fields, split them into str fields or list type fields as per description dataframe
    Args:
        ontological_fields:
        desc_df:

    Returns:

    """
    curated_fields = desc_df[desc_df['column_name'].str.startswith('curated')]
    schema_dict = dict(zip(curated_fields['column_name'], curated_fields['column_type']))
    list_fields = [k for k, v in schema_dict.items() if v == 'array<string>' and k in ontological_fields]
    str_fields = [k for k, v in schema_dict.items() if v == 'string' and k in ontological_fields]
    return str_fields, list_fields


def schema_correction(meta, err):
    """
    Type Modification:
        - If errors are found, collect the errors and perform in-memory modifications of the metadata values.
        This correction can be one of the following:
            If the value should be a list but the incoming value is a string, typecast it into a list
            E.g. ‘Cancer’ will be changed to ['Cancer']

            If the value should be a string, but the incoming value is an integer, typecast it into a string.
            E.g. 2021 will change to ‘2021’.
        - If no errors are found, perform no in-memory modifications.
    Args:
        meta: sample level metadata dict
        err: dict of schema errors

    Returns:

    """
    for i in range(len(err)):
        field = err[i]['loc'][0]
        if err[i]['type'] == 'type_error.list':
            meta[field] = [meta[field]]
        elif err[i]['type'] == 'type_error.str':
            meta[field] = str(meta[field])
    return meta
