import sys
import traceback as tb

import pandas as pd
from pydantic import ValidationError
from tabulate import tabulate

from ipc.utility.helper import schema_correction
from ipc.validator.schemas.dataset_metadata_schemas import build_schema_for_repo


def check_metadata_for_errors(repo, desc_df, metadata, level='value', print_table=False):
    """
    Utility to return errors for a single sample level metadata dict
    Args:
        desc_df:
        metadata: dict containing the sample level metadata to be checked
        level: level of checks to do, level='schema', 'value'
        print_table: flag to print the results as a table or not

    Returns:
        errors and desc :  errors as an error df
        status (boolean): Status for if the given metadata passed all checks or not
    """
    error_df = pd.DataFrame(columns=['Field', 'Error Message', 'Error Type', 'Repo',
                                     'dataset_id', 'key'])
    additional_errors = pd.DataFrame(columns=['Field', 'Error Message', 'Error Type', 'Repo',
                                              'dataset_id', 'key'])
    collected_errors = pd.DataFrame(columns=['Field', 'Error Message', 'Error Type', 'Repo',
                                             'dataset_id', 'key'])
    try:
        schema, schema_with_validators = build_schema_for_repo(desc_df)
        if not schema or not schema_with_validators:
            raise Exception(f'Error building pydantic classes. {tb.format_exc()}')
        if level == 'schema':
            try:
                schema(**metadata)
            except ValidationError as schema_errors:
                error_df = pd.DataFrame(schema_errors.errors())
                # Rename columns to our desired names
            if not error_df.empty:
                error_df.rename(columns={'loc': 'Field',
                                         'msg': 'Error Message',
                                         'type': 'Error Type'}, inplace=True)
                error_df['Repo'] = repo
                if 'dataset_id' in metadata:
                    if isinstance(metadata['dataset_id'], list):
                        error_df['dataset_id'] = metadata['dataset_id'][0]
                    else:
                        error_df['dataset_id'] = metadata['dataset_id']
                else:
                    error_df['dataset_id'] = 'NA'
                error_df['key'] = metadata['key']
                error_df['Field'] = error_df['Field'].apply(lambda row: row[0])
                error_df.drop_duplicates(inplace=True)
                if print_table:
                    # 'print_table' is True: returns formatted table for the command line
                    error_df['Error Message'] = error_df['Error Message'].str.wrap(20)
                    print(tabulate(error_df, headers="keys", tablefmt="fancy_grid"))
                return error_df, False
        elif level == 'value':
            """Step 1. Check only for schema"""
            try:
                schema(**metadata)
            except ValidationError as schema_errors:
                error_df = pd.DataFrame(schema_errors.errors())
                # Rename columns to our desired names
                error_df.rename(columns={'loc': 'Field',
                                         'msg': 'Error Message',
                                         'type': 'Error Type'}, inplace=True)
                errors_to_correct = schema_errors.errors()
                # Perform intermediate in-memory corrections
                if errors_to_correct:
                    metadata = schema_correction(metadata, errors_to_correct)

            """Step 2. Check for value checks with corrected schema"""
            try:
                schema_with_validators(**metadata)
            except ValidationError as value_errors:
                additional_errors = pd.DataFrame(value_errors.errors())
                # Rename columns to our desired names
                additional_errors.rename(columns={'loc': 'Field',
                                                  'msg': 'Error Message',
                                                  'type': 'Error Type'}, inplace=True)
            # If some errors are found in any of the two checks. Concatenate them together.
            if not error_df.empty and not additional_errors.empty:
                collected_errors = pd.concat([error_df, additional_errors], ignore_index=True)
                collected_errors.drop_duplicates(inplace=True)
            elif not error_df.empty and additional_errors.empty:
                collected_errors = error_df.copy()
            elif error_df.empty and not additional_errors.empty:
                collected_errors = additional_errors.copy()
            collected_errors['Repo'] = repo
            if 'dataset_id' in metadata:
                if isinstance(metadata['dataset_id'], list):
                    collected_errors['dataset_id'] = metadata['dataset_id'][0]
                else:
                    collected_errors['dataset_id'] = metadata['dataset_id']
            else:
                collected_errors['dataset_id'] = 'NA'
            collected_errors['key'] = metadata['key']
            collected_errors['Field'] = collected_errors['Field'].apply(lambda row: row[0])
            if print_table:
                # 'print_table' is True: returns formatted table for the command line
                collected_errors['Error Message'] = collected_errors['Error Message'].str.wrap(20)
                print(tabulate(collected_errors, headers="keys", tablefmt="fancy_grid"))
            return collected_errors, False
        else:
            raise Exception(f'Invalid value for argument "level": {level} ')
    except Exception as e:
        print(tb.format_exc())
        exception_type, exception_object, exception_traceback = sys.exc_info()
        filename = exception_traceback.tb_frame.f_code.co_filename
        line_number = exception_traceback.tb_lineno
        print("Exception type: ", exception_type)
        print("File name: ", filename)
        print("Line number: ", line_number)
        return pd.DataFrame(), False
    return collected_errors, True
